// This software is supplied under the terms of a license agreement or
// nondisclosure agreement with B-Alive GmbH and may not 
// be copied or disclosed except in accordance with the terms of that 
// agreement.
//
//      Copyright (c) 2004-2008 B-Alive GmbH.
//      All Rights Reserved.
//
// B-Alive GmbH
// http://www.b-alive.de
//
// author: ank
// date: 19.07.2006

#ifndef SKINNING_H
#define SKINNING_H

#include <ShaderFragments.hlsl>
#include <shadowmap.hlsl>

// ================================================================================================
// function for fog calculations @ vertex shader
// p_iMethode = fogging methode, 0 = linear, 1 = exp, 2 = exp2
// ================================================================================================
float CalculateFogValue(const int p_iMethode, float p_fZDepth)
{
    float fFogValue = 0;
    
    if (p_iMethode == 0)
        fFogValue = (FogEnd - p_fZDepth) / (FogEnd - FogStart);
    else if (p_iMethode == 1)
        fFogValue = exp(-FogDensity * p_fZDepth);
    else if (p_iMethode == 0)
        fFogValue = exp(-pow(FogDensity * p_fZDepth,2));
        
    return fFogValue;
}

// ================================================================================================
// final function for mixing in pixel shader
// ================================================================================================
float4 MixColorWithFog(float4 p_vColor, float p_fFogValue)
{
    return lerp(FogColor, p_vColor, p_fFogValue);
}

// ================================================================================================
float4 BlinnPhong(float3 normal, float3 lightDir, float specPower)
{
    float  nDotL   = dot(normal, -lightDir);
    float3 halfVec = normalize(-lightDir + 2*normal * nDotL);
    float  nDotH   = dot(normal, halfVec);
    
    float4 shading = pow(max(0, nDotH), specPower) * max(0,nDotL) * g_fGlossiness;
    return shading;
}

// ================================================================================================
float CalcDiffuseShading(float3 vNormal, float3 vLightDir, float fShadow, out float fNDotL)
{
#if !defined(DIFFUSE)
    
    fNDotL = dot(vNormal, -vLightDir);
    return 0;
    
#else
#   if !defined(HALF_LAMBERT_SHADING)
    
    fNDotL = dot(vNormal, -vLightDir);
    return saturate(fNDotL * fShadow); // needs 1 instruction less than max
    //return max(0, fNDotL * fShadow);
    
    //return max(0, dot(vNormal, -vLightDir));
    //return min(dot(vNormal, -vLightDir) * 0.5, fShadow) + 0.5;

#   else // HALF_LAMBERT_SHADING == defined
    
    // half lambert shading for softer illumination (in shadowed area)
    fNDotL = dot(vNormal, -vLightDir);
    //float f = max(0, fNDotL * 0.5 + 0.5 * fShadow);
    float f = saturate(fNDotL * 0.5 + 0.5 * fShadow); // needs 1 instruction less than max
    f = f * f;
    
    return  f;
    
#   endif // HALF_LAMBERT_SHADING
#endif // DIFFUSE
}

// ================================================================================================
float CalcSpecularShading(float3 vReflect, float3 vViewDir)
{
#if !defined(SPECULAR)
    return 0;
#else
    return pow(max(0, dot(vReflect, -vViewDir)), MaterialPower/4) * g_fGlossiness;
#endif
}

// ================================================================================================
float3 CalcReflection(float3 vNormal, float3 vLightDir)
{
    return normalize(reflect(vLightDir, vNormal));
}

// ================================================================================================
float4 CalcAmbient()
{
    return MaterialAmbient * LightAmbient;
}

// ================================================================================================
float4 CalcDiffuse(float3 vNormal, float3 vLightDir)
{
    //return MaterialDiffuse * LightDiffuse * max(0, dot(vNormal, -vLightDir));
    float f = 0.0f;
    return MaterialDiffuse * LightDiffuse * CalcDiffuseShading(vNormal, vLightDir, 1, f);
}

// ================================================================================================
// Specular wird draufaddiert -> Alpha auf 0 setzen!
// ================================================================================================
float4 CalcSpecular(float3 vNormal, float3 vLightDir, float3 vViewDir)
{
    float3 r = CalcReflection(vNormal, vLightDir);
    //float4 col = MaterialSpecular * LightSpecular * pow(max(0, dot(r, -vViewDir)), MaterialPower/4) * g_fGlossiness;
    float4 col = MaterialSpecular * LightSpecular * CalcSpecularShading(r, vViewDir);
    col.a = 0;
    return col;
}

// ================================================================================================
float4 CalcSpecularWithAlpha(float3 vNormal, float3 vLightDir, float3 vViewDir)
{
    float3 r = CalcReflection(vNormal, vLightDir);
    //float4 col = MaterialSpecular * LightSpecular * pow(max(0, dot(r, -vViewDir)), MaterialPower/4) * g_fGlossiness;
    float4 col = MaterialSpecular * LightSpecular * CalcSpecularShading(r, vViewDir); //pow(max(0, dot(r, -vViewDir)), MaterialPower/4) * g_fGlossiness;
    return col;
}

// ================================================================================================
// By mad: optimized version to seperate calculation (saves 1x dot(N,L) - only sign has to be switched!)
// ================================================================================================
void CalcDiffuseAndSpecular(float3 vNormal, float3 vLightDir, float3 vViewDir, out float4 diffuse, out float4 specular, float fShadow)
{
    //float NdotL = dot(vNormal, -vLightDir) * fShadow; // kill lighting...
    float NdotL = 0;
    float shading = CalcDiffuseShading(vNormal, vLightDir, fShadow, NdotL); // kill lighting...
    
    // diffuse color:
    //diffuse = MaterialDiffuse * LightDiffuse * max(0,NdotL);
    diffuse = MaterialDiffuse * LightDiffuse * shading;
    
    // specular color | first reflect vector
    float3 r = normalize(vLightDir + 2*vNormal * NdotL);
        
    //specular = MaterialSpecular * LightSpecular * pow(max(0, dot(r, -vViewDir)), MaterialPower/4) * g_fGlossiness;
    specular = MaterialSpecular * LightSpecular * CalcSpecularShading(r, vViewDir) * fShadow;
    
    //specular.a = 0;
}

// ================================================================================================
// Lerpfunktion - HLSL-lerp unterhalb von ps1.4 erzeugt zuviele Instructions
// ================================================================================================
float4 float4Lerp(float4 vSrc, float4 vDst, float fFac)
{
    return vSrc + (vDst - vSrc) * fFac;
}

// ================================================================================================
float4 float4LerpXYZ(float4 vSrc, float4 vDst, float fFac)
{
    return float4(vSrc.rgb + (vDst.rgb - vSrc.rgb) * fFac, vSrc.w);
}

// ================================================================================================
float CalculateWaterFogFX(float p_vWorldPositionZ)
{
    return (WaterHeight - p_vWorldPositionZ) * DeepSeaFactor;
}

// ================================================================================================
float CalculateCausticStrength(float3 WorldNormal)
{
    return max(0.1, dot(WorldNormal, float3(0,0,1)));
}

// ================================================================================================
float3 GetCausticColor(float p_fUnderwaterFactor, float2 p_vTexCoords, float p_fSpecFactor)
{
    float4 colCaust = tex2D(SamplerCaustics, p_vTexCoords * 2.0);
    colCaust *= p_fSpecFactor * min(1, p_fUnderwaterFactor * 6);
    return colCaust.rgb * CaustEnhanceFac;
}

// ================================================================================================
float4 ModulateSrcAlphaInvSrcAlpha(float4 p_ColSrc, float4 p_ColDst)
{
    float4 col1 = float4(p_ColSrc.rgb * p_ColSrc.a, p_ColSrc.a);    // src alpha
    float4 col2 = float4(p_ColDst * (1-p_ColSrc.a));                // inv src alpha
    return col1 + col2;
}

// ================================================================================================
struct VS_OutCol
{    
    float4 Pos  : POSITION;
    //float4 Col  : COLOR;
    float4 Amb  : COLOR;
    float4 Diff : COLOR1;

#if defined(ENABLE_SHADOWMAPPING)
    float4 LightVPS : TEXCOORD0;
#endif

};

// ================================================================================================
struct VS_OutColTex
{    
    float4 Pos  : POSITION;
    //float4 Col  : COLOR;
    float4 Amb  : COLOR;
    float4 Diff : COLOR1;
    float2 Tex  : TEXCOORD0;

#if defined(ENABLE_SHADOWMAPPING)
    float4 LightVPS : TEXCOORD1;
#endif
};

// ================================================================================================
struct VS_OutColSpec
{
    float4 Pos  : POSITION;
    //float4 Col  : COLOR;
    float4 Amb  : COLOR;
    float4 Diff : COLOR1;
    float4 Spec : TEXCOORD0;

#if defined(ENABLE_SHADOWMAPPING)
    float4 LightVPS : TEXCOORD1;
#endif
};

// ================================================================================================
struct VS_OutColSpec4Tex
{
    float4 Pos  : POSITION;
    float4 Amb  : COLOR;
    float4 Diff : COLOR1;
    float4 Spec : TEXCOORD0;
    float4 Tex01  : TEXCOORD1;
    float4 Tex23  : TEXCOORD2;

#if defined(ENABLE_SHADOWMAPPING)
    float4 LightVPS : TEXCOORD3;
#endif
};

struct VS_OutColSpec4TexWaterFog
{
	float4 Pos  : POSITION;
    float4 Amb  : COLOR;
    float4 Diff : COLOR1;
    float4 Spec : TEXCOORD0;
    float4 Tex01  : TEXCOORD1;
    float4 Tex23  : TEXCOORD2;
	float4 NormalAndWaterFog : TEXCOORD3;	// xyz = normal, w = WaterFog

#if defined(USE_DEPTH_MAP_MRT)
	float4 PosVS : TEXCOORD4;		// xyz = VS position, w = Projected depth
#if defined(ENABLE_SHADOWMAPPING)
    float4 LightVPS : TEXCOORD5;
#endif
#else
#if defined(ENABLE_SHADOWMAPPING)
    float4 LightVPS : TEXCOORD4;
#endif
#endif
};

// ================================================================================================
// Vertex shader output / pixel shader input with texture space data
// ================================================================================================
struct VS_OutTSD // TSD = texture space data
{    
    float4 Pos          : POSITION;
    float4 Col          : COLOR0;        // ambient light
    float4 Spec         : COLOR1;        // specular light
    float4 Tex01        : TEXCOORD0;    // texture set 0 & 1
    float4 Tex23        : TEXCOORD1;    // texture set 2 & 3
    float3 TSLightDir   : TEXCOORD2;    // Texture-/TangentSpace light direction
    float3 TSViewDir    : TEXCOORD3;
    float4 WorldUpAndWater : TEXCOORD4; // first 3 Values = WorldUp-Vec, W = WaterFog-Param

#if defined(USE_DEPTH_MAP_MRT)
	float4 PosVS		: TEXCOORD5;	// xyz = Position in view space, w = Projected Position
#if defined(ENABLE_SHADOWMAPPING)
    float4 LightVPS     : TEXCOORD6;    // by mad: Position in LightSpace
#endif
#else // USE_DEPTH_MAP_MRT
#if defined(ENABLE_SHADOWMAPPING)
    float4 LightVPS     : TEXCOORD5;    // by mad: Position in LightSpace
#endif
#endif // USE_DEPTH_MAP_MRT
};

// ================================================================================================
// Simple geometry data output
// ================================================================================================
struct VS_OutNormal
{
    float3 Pos      : POSITION;
    float3 Normal   : NORMAL;
};

// ================================================================================================
struct VS_OutNormalTex
{
	float3 Pos : POSITION;
	float3 Normal : NORMAL;
	float2 Tex : TEXCOORD0;
};

// ================================================================================================
// Skinning stuff
// ================================================================================================

// ================================================================================================
struct VS_InSkinning
{
    float4  Pos             : POSITION;
    float4  BlendWeights    : BLENDWEIGHT;
    float4  BlendIndices    : BLENDINDICES;
    float3  Normal          : NORMAL;
    float2  Tex             : TEXCOORD0;
};

// ================================================================================================
struct VS_InSkinning_4Tex
{
    float4  Pos             : POSITION;
	float4  Color			: COLOR;		
    float4  BlendWeights    : BLENDWEIGHT;
    float4  BlendIndices    : BLENDINDICES;
    float3  Normal          : NORMAL;
    float2  Tex             : TEXCOORD0;
    float2  Tex1            : TEXCOORD1;
    float2  Tex2            : TEXCOORD2;
    float2  Tex3            : TEXCOORD3;
};

// ================================================================================================
VS_OutNormal Skinning(float4  Pos           : POSITION,
                      float4  BlendWeights  : BLENDWEIGHT,
                      float4  BlendIndices  : BLENDINDICES,
                      float3  Normal        : NORMAL,
                      uniform int NumBones)
{
    VS_OutNormal o;
    o.Pos = 0;
    o.Normal = 0;
    float LastWeight = 0;

    // Compensate for lack of UBYTE4 on Geforce3
    int4 IndexVector = D3DCOLORtoUBYTE4(BlendIndices);

    // cast the vectors to arrays for use in the for loop below
    float BlendWeightsArray[4] = (float[4])BlendWeights;
    int   IndexArray[4]        = (int[4])IndexVector;

    // calculate the pos/normal using the "normal" weights and accumulate the weights to calculate the last weight
    for (int iBone=0; iBone<NumBones-1; iBone++)
    {
        LastWeight = LastWeight + BlendWeightsArray[iBone];
        o.Pos += mul(Pos, matWorldArray[IndexArray[iBone]]) * BlendWeightsArray[iBone];
        o.Normal += mul(Normal, (float3x3)matWorldArray[IndexArray[iBone]]) * BlendWeightsArray[iBone];
    }
    LastWeight = 1 - LastWeight; 

    // Now that we have the calculated weight, add in the final influence
    o.Pos += mul(Pos, matWorldArray[IndexArray[NumBones-1]]) * LastWeight;
    o.Normal += mul(Normal, matWorldArray[IndexArray[NumBones-1]]) * LastWeight;

    // normalize normal
    o.Normal = normalize(o.Normal);

    return o;
}

// ================================================================================================
float3 SkinningPosOnly(float4  Pos           : POSITION,
                       float4  BlendWeights  : BLENDWEIGHT,
                       float4  BlendIndices  : BLENDINDICES,
                       uniform int NumBones) : POSITION
{
    float3 NewPos = 0;
    float LastWeight = 0;

    // Compensate for lack of UBYTE4 on Geforce3
    int4 IndexVector = D3DCOLORtoUBYTE4(BlendIndices);

    // cast the vectors to arrays for use in the for loop below
    float BlendWeightsArray[4] = (float[4])BlendWeights;
    int   IndexArray[4]        = (int[4])IndexVector;

    // calculate the pos using the "normal" weights and accumulate the weights to calculate the last weight
    for (int iBone=0; iBone<NumBones-1; iBone++)
    {
        LastWeight = LastWeight + BlendWeightsArray[iBone];
        NewPos += mul(Pos, matWorldArray[IndexArray[iBone]]) * BlendWeightsArray[iBone];
    }
    LastWeight = 1.0f - LastWeight; 

    // Now that we have the calculated weight, add in the final influence
    NewPos += mul(Pos, matWorldArray[IndexArray[NumBones-1]]) * LastWeight;

    return NewPos;
}


// ================================================================================================
// NBT skinning stuff (normal binormal tangent)
// ================================================================================================
struct VS_InNBT
{
    float4  Pos         : POSITION;
    float3  Normal      : NORMAL;
    float3  Binormal    : BINORMAL;
    float3  Tangent     : TANGENT;
    float2  Tex0        : TEXCOORD0;
    float2  Tex1        : TEXCOORD1;
    float2  Tex2        : TEXCOORD2;
    float2  Tex3        : TEXCOORD3;
};

// ================================================================================================
struct VS_InSkinnedNBT
{
    float4  Pos             : POSITION;
    float4  BlendWeights    : BLENDWEIGHT;
    float4  BlendIndices    : BLENDINDICES;
    float3  Normal          : NORMAL;
    float3  Binormal        : BINORMAL;
    float3  Tangent         : TANGENT;
    float2  Tex0            : TEXCOORD0;
    float2  Tex1            : TEXCOORD1;
    float2  Tex2            : TEXCOORD2;
    float2  Tex3            : TEXCOORD3;
};

// ================================================================================================
struct VS_OutSkinnedNBT
{
    float3 Pos      : POSITION;
    float3 Normal   : NORMAL;
    float3 Binormal : BINORMAL;
    float3 Tangent  : TANGENT;
};

// ================================================================================================
// All normals have to be transformed. Don't use the crossproduct to get a normal -> yields to
// unexpected results.
// ================================================================================================
VS_OutSkinnedNBT SkinningNBT(VS_InSkinnedNBT i, uniform int NumBones)
{
    VS_OutSkinnedNBT o;
    o.Pos = 0;
    o.Normal = 0;
    o.Tangent = 0;
    o.Binormal = 0;
    float LastWeight = 0;

    // Compensate for lack of UBYTE4 on Geforce3
    int4 IndexVector = D3DCOLORtoUBYTE4(i.BlendIndices);

    // cast the vectors to arrays for use in the for loop below
    float BlendWeightsArray[4] = (float[4])i.BlendWeights;
    int   IndexArray[4]        = (int[4])IndexVector;

    // calculate the pos/normal using the "normal" weights and accumulate the weights to calculate the last weight
    for (int iBone=0; iBone<NumBones-1; iBone++)
    {
        LastWeight = LastWeight + BlendWeightsArray[iBone];
        o.Pos += mul(i.Pos, matWorldArray[IndexArray[iBone]]) * BlendWeightsArray[iBone];
        o.Normal += mul(i.Normal, (float3x3)matWorldArray[IndexArray[iBone]]) * BlendWeightsArray[iBone];
        o.Tangent += mul(i.Tangent, (float3x3)matWorldArray[IndexArray[iBone]]) * BlendWeightsArray[iBone];
        o.Binormal += mul(i.Binormal, (float3x3)matWorldArray[IndexArray[iBone]]) * BlendWeightsArray[iBone];
    }
    LastWeight = 1.0f - LastWeight; 

    // Now that we have the calculated weight, add in the final influence
    o.Pos += mul(i.Pos, matWorldArray[IndexArray[NumBones-1]]) * LastWeight;
    o.Normal += mul(i.Normal, (float3x3)matWorldArray[IndexArray[NumBones-1]]) * LastWeight;
    o.Tangent += mul(i.Tangent, (float3x3)matWorldArray[IndexArray[NumBones-1]]) * LastWeight;
    o.Binormal += mul(i.Binormal, (float3x3)matWorldArray[IndexArray[NumBones-1]]) * LastWeight;

    // normalize normals
    o.Normal = normalize(o.Normal);
    o.Tangent = normalize(o.Tangent);
    o.Binormal = normalize(o.Binormal);

    return o;
}

// ================================================================================================
// Transform only position. Normals are not needed for shadow stuff.
// ================================================================================================
float3 SkinningNBTPosOnly(VS_InSkinnedNBT i, uniform int NumBones) : POSITION
{
    float3 Pos = 0;
    float LastWeight = 0;

    // Compensate for lack of UBYTE4 on Geforce3
    int4 IndexVector = D3DCOLORtoUBYTE4(i.BlendIndices);

    // cast the vectors to arrays for use in the for loop below
    float BlendWeightsArray[4] = (float[4])i.BlendWeights;
    int   IndexArray[4]        = (int[4])IndexVector;

    // calculate the pos/normal using the "normal" weights and accumulate the weights to calculate the last weight
    for (int iBone=0; iBone<NumBones-1; iBone++)
    {
        LastWeight = LastWeight + BlendWeightsArray[iBone];
        Pos += mul(i.Pos, matWorldArray[IndexArray[iBone]]) * BlendWeightsArray[iBone];
    }
    LastWeight = 1.0f - LastWeight; 

    // Now that we have the calculated weight, add in the final influence
    Pos += mul(i.Pos, matWorldArray[IndexArray[NumBones-1]]) * LastWeight;

    return Pos;
}


// ================================================================================================
void UpdateSnowLayer(inout float4 vDiffuseColor, inout float4 vSpecularColor, in float3 vNormal, in float3 vWorldUp, in float2 vTexCoord)
{
    // schneelayer testweise mal:
    //*
    if (g_fSnowCoverage > 0.f)
    {
        float fangle = max(0, dot(vNormal, vWorldUp));
        // keep some snow on larger angles
		fangle = fangle + g_fSnowCoverage + g_fSnowConstFactor;    // 0.75*cover - 0.75 -> done in program code?!
		fangle = saturate(fangle) * vDiffuseColor.a;
        float fInvAngle = (1-fangle);
#if !defined(ADVANCED_SHADING)
        vDiffuseColor.xyz = vDiffuseColor*fInvAngle + float3(0.65,0.67,0.75)*fangle;
#else // ADVANCED_SHADING = 1
        vDiffuseColor.xyz = vDiffuseColor*fInvAngle + float3(0.65,0.67,0.75)*tex2D(SamplerSnow, vTexCoord)*fangle;
#endif // ADVANCED_SHADING
        // todo: modify specular?!...
        vSpecularColor *= fInvAngle;
    }
    //*/
}

// ================================================================================================
float calculateSnowFactor(in float3 vNormal)
{
	float angle = max(g_fSnowConstFactor, dot(vNormal, float3(0.0f, 0.0f, 1.0f)));
	return saturate(angle * g_fSnowCoverage);
}

// ================================================================================================
// Pixel shader output with texture space data (normal mapping) based lighting.
// Requires PS 2.0 or above!
// ================================================================================================
PS_Out PS_TSDLighting(VS_OutTSD i, uniform bool bSpecAlpha, uniform bool bCalcWaterFX)
{
	PS_Out output = (PS_Out)0;

#if defined(ENABLE_SHADOWMAPPING)
    // by mad: um schatten erweitert
    float fCurrentDepth = GetDepthInLightSpace(i.LightVPS);
    float fShadow = GetObjectInShadowFactor(fCurrentDepth, i.LightVPS, ShadowMap);
#else
    float fShadow = 1.f;
#endif
      
#if defined(ALBEDO)
    float4 colTex = tex2D(SamplerBase, i.Tex01.xy) * ColorBase;
#else
    float4 colTex = ColorBase;
#endif

    float3 colNormal = tex2D(Sampler0, i.Tex01.xy) * 2 - 1; // "decompress" normal
    
    // float4 colDif = CalcDiffuse(colNormal, i.TSLightDir);
    // float4 colSpec = CalcSpecular(colNormal, i.TSLightDir, i.TSViewDir);
    float4 colDif, colSpec;
    CalcDiffuseAndSpecular(colNormal, i.TSLightDir, i.TSViewDir, colDif, colSpec, fShadow);

    UpdateSnowLayer(colTex, colSpec, colNormal, i.WorldUpAndWater.xyz, i.Tex01.xy);

#if defined(ENABLE_SHADOWMAPPING)
    colTex.xyz = colTex.xyz *(i.Col + colDif) + colSpec;
	if (bSpecAlpha)
		colTex.a += fShadow*colSpec.a*colTex.a;
    
    if (bCalcWaterFX)
    {
        i.WorldUpAndWater.w = saturate(i.WorldUpAndWater.w);
        colTex.rgb += GetCausticColor(i.WorldUpAndWater.w, i.Tex01.xy, i.Spec.w);// * fShadow);
        colTex = float4LerpXYZ(colTex, WATERGROUND_COLOR, i.WorldUpAndWater.w);
    }
#else
    colTex.xyz = colTex.xyz * (i.Col + colDif) + colSpec;
	if (bSpecAlpha)
		colTex.a += colSpec.a*colTex.a;
	
	if (bCalcWaterFX)
    {
        i.WorldUpAndWater.w = saturate(i.WorldUpAndWater.w);
	    colTex.rgb += GetCausticColor(i.WorldUpAndWater.w, i.Tex01.xy, i.Spec.w);
        colTex = float4LerpXYZ(colTex, WATERGROUND_COLOR, i.WorldUpAndWater.w);
    }
#endif
    
	output.Color = colTex;

#if defined(USE_DEPTH_MAP_MRT)
	output.Depth.x = 1.0f - i.PosVS.w;
	output.Depth.y = i.PosVS.z;
	output.Depth.a = 1.0f;
#endif

    //float4 col = colTex * (i.Col + colDif) * ColorBase + colSpec;
    //col = float4Lerp(colTex * i.Col + i.Spec, col, g_fNBTValue); // fr Lerp zu VertexLighting
    return output;
}

// ================================================================================================
PS_Out PS_TSDLighting_3Tex(VS_OutTSD i, bool bSecondColorBlend, uniform bool bSpecAlpha, uniform bool bCalcWaterFX)
{
	PS_Out output = (PS_Out)0;

#if defined(ENABLE_SHADOWMAPPING)
    // by mad: hier Schatten versuchen zu verarbeiten:
    float fCurrentDepth = GetDepthInLightSpace(i.LightVPS);
    float fShadow = GetObjectInShadowFactor(fCurrentDepth, i.LightVPS, ShadowMap);
#else
    float fShadow = 1.f;
#endif    
    
#if defined(ALBEDO)
    float4 col = tex2D(SamplerBase, i.Tex01.xy);
#else
    float4 col = 1;
#endif
    float3 colNormal = tex2D(Sampler0, i.Tex01.xy) * 2 - 1; // "decompress" normal
    float4 col2 = tex2D(Sampler2, i.Tex01.zw);
    //float4 col3 = tex2D(Sampler3, i.Tex23.xy);

    //////////////////////////////
    //fShadow = 1.f;
    //return float4(i.TSLightDir, 1);
    //return float4(i.TSViewDir, 1);
    //////////////////////////////
    
    //float4 colDif = CalcDiffuse(colNormal, i.TSLightDir);
    //float4 colSpec = CalcSpecular(colNormal, i.TSLightDir, i.TSViewDir);
    float4 colDif, colSpec;
    CalcDiffuseAndSpecular(colNormal, i.TSLightDir, i.TSViewDir, colDif, colSpec, fShadow);

    //////////////////////////////
    //return colDif;
    //////////////////////////////
    
    if (bSecondColorBlend)
    {
        float4 col2 =  tex2D(Sampler4, i.Tex01.xy);
        float f = (1.0 - g_fSecondColorBlend);
        col.xyz = col * f + g_fSecondColorBlend * col2;
        col.a   = ((col2.a + 0.001) >= g_fSecondColorDecay) ? col.a : 0;
        
        // with dehydration/colorblend less specular [only when without waterfx - cause of shader limits]
        if (!bCalcWaterFX)
            colSpec *= f;
    }
        
    col  *= ColorBase;
    col2 *= ColorLayer0;
    //col3 *= ColorLayer1;

    col.xyz = ModulateSrcAlphaInvSrcAlpha(col2, col);
    //col.xyz = ModulateSrcAlphaInvSrcAlpha(col3, col);
    
    //UpdateSnowLayer(col, colSpec, colNormal, i.WorldUpAndWater.xyz, i.Tex01.xy);
	float snowFactor = calculateSnowFactor(colNormal);
	col.xyz = float4LerpXYZ(col, float4(1.0f, 1.0f, 1.0f, 1.0f), snowFactor);
    
#if defined(ENABLE_SHADOWMAPPING)
    col.xyz = col.xyz *(i.Col + colDif) + colSpec;
	if (bSpecAlpha)
		col.a += fShadow*colSpec.a*col.a;

    if (bCalcWaterFX)
    {
        i.WorldUpAndWater.w = saturate(i.WorldUpAndWater.w);
	    col.rgb += GetCausticColor(i.WorldUpAndWater.w, i.Tex01.xy, i.Spec.w); //*fShadow);
	    col = float4LerpXYZ(col, WATERGROUND_COLOR, i.WorldUpAndWater.w);
	}
#else
    col.xyz = col.xyz * (i.Col + colDif) + colSpec;
	if (bSpecAlpha)
		col.a += colSpec.a*col.a;
	
    if (bCalcWaterFX)
    {
        i.WorldUpAndWater.w = saturate(i.WorldUpAndWater.w);
        col.rgb += GetCausticColor(i.WorldUpAndWater.w, i.Tex01.xy, i.Spec.w);
        col = float4LerpXYZ(col, WATERGROUND_COLOR, i.WorldUpAndWater.w);
    }
#endif

	output.Color = col;

#if defined(USE_DEPTH_MAP_MRT)
	output.Depth.x = 1.0f - i.PosVS.w;
	output.Depth.y = i.PosVS.z;
	output.Depth.a = 1.0f;
#endif

    return output;
    // return col * (i.Col + colDif) * ColorBase + colSpec;
}

// ================================================================================================
PS_Out PS_VertexLighting_NoTex(VS_OutColSpec p_psIn)
{
	PS_Out output = (PS_Out)0;

#if defined(ENABLE_SHADOWMAPPING)
    // by mad: um schatten erweitert
    float fCurrentDepth = GetDepthInLightSpace(p_psIn.LightVPS);
    float fShadow = GetObjectInShadowFactor(fCurrentDepth, p_psIn.LightVPS, ShadowMap);
    
	output.Color = saturate(p_psIn.Amb + fShadow * p_psIn.Diff) + fShadow * saturate(p_psIn.Spec);
#else
    output.Color = saturate(p_psIn.Amb + p_psIn.Diff) + saturate(p_psIn.Spec);
#endif

	return output;
}

// ================================================================================================
PS_Out PS_VertexLighting(VS_OutColSpec4Tex p_psIn, uniform bool bSecondColorBlend = false)
{
	PS_Out output = (PS_Out)0;

#if defined(ENABLE_SHADOWMAPPING)
    // by mad: um schatten erweitert
    float fCurrentDepth = GetDepthInLightSpace(p_psIn.LightVPS);
    float fShadow = GetObjectInShadowFactor(fCurrentDepth, p_psIn.LightVPS, ShadowMap);

    float4 col = tex2D(SamplerBase, p_psIn.Tex01.xy);
    
    float4 colSpec = saturate(p_psIn.Spec);
    if (bSecondColorBlend)
    {
        float4 col2 =  tex2D(Sampler4, p_psIn.Tex01.xy);
        float f = (1.0 - g_fSecondColorBlend);
        col.xyz = col * f + g_fSecondColorBlend * col2;
        col.a   = ((col2.a + 0.001) >= g_fSecondColorDecay) ? col.a : 0;
        
        // with dehydration/colorblend less specular
        colSpec *= f;
    }
    
    output.Color = col * saturate(p_psIn.Amb + fShadow*p_psIn.Diff) + fShadow*colSpec;
#else
    float4 col = tex2D(SamplerBase, p_psIn.Tex01.xy);
    
    float4 colSpec = saturate(p_psIn.Spec);
    if (bSecondColorBlend)
    {
        float4 col2 =  tex2D(Sampler4, p_psIn.Tex01.xy);
        float f = (1.0 - g_fSecondColorBlend);
        col.xyz = col * f + g_fSecondColorBlend * col2;
        col.a   = ((col2.a + 0.001) >= g_fSecondColorDecay) ? col.a : 0;
        
        // with dehydration/colorblend less specular
        colSpec *= f;
    }
	
	output.Color = col * saturate(p_psIn.Amb + p_psIn.Diff) + colSpec;
    // return col * p_psIn.Col + p_psIn.Spec;
#endif
	
	return output;
}

PS_Out PS_VertexLighting_3Tex(VS_OutColSpec4TexWaterFog p_psIn, uniform bool bSecondColorBlend = false, uniform bool bUnderWaterFx = false, uniform bool bUseShadowDepthOffset = false)
{
	PS_Out output = (PS_Out)0;

    float4 col = tex2D(SamplerBase, p_psIn.Tex01.xy);
    float4 col2 = tex2D(Sampler2, p_psIn.Tex01.zw);
    float4 col3 = tex2D(Sampler3, p_psIn.Tex23.xy);
    //float4 col4 = tex2D(Sampler4, p_psIn.Tex23.zw);

    float4 colSpec = saturate(p_psIn.Spec);
    if (bSecondColorBlend)
    {
        float4 col2 =  tex2D(Sampler4, p_psIn.Tex01.xy);
        float f = (1.0 - g_fSecondColorBlend);
        col.xyz = col * f + g_fSecondColorBlend * col2;
        col.a   = ((col2.a + 0.001) >= g_fSecondColorDecay) ? col.a : 0;
        
        // with dehydration/colorblend less specular
        colSpec *= f;
    }

    col *= ColorBase;
    col2 *= ColorLayer0;
    col3 *= ColorLayer1;
    //col4 *= ColorLayer2;
    
    col.xyz = ModulateSrcAlphaInvSrcAlpha(col2, col);
    col.xyz = ModulateSrcAlphaInvSrcAlpha(col3, col);
    //col = ModulateSrcAlphaInvSrcAlpha(col4, col);
	
	// by jow: aus dem vertex shader nach hier bernommen, damit diese Berechnung fr den Schnee unabhngig von der Lichtberechnung ist
	col.rgb *= ColorBase.rgb;

	float snowFactor = calculateSnowFactor(p_psIn.NormalAndWaterFog.xyz);
	col.xyz = float4LerpXYZ(col, float4(1.0f, 1.0f, 1.0f, 1.0f), snowFactor);

#if defined(ENABLE_SHADOWMAPPING)
	float fShadow;
	// Spezialfall: Fr Skinning wird in p_psIn.LightVPS die Worldspace Position bergeben, ansonsten die Lightspace Position
	if(bUseShadowDepthOffset)
	{
		// Anstelle eines festen Depthbias zu verwenden, wird ein Offset in Worldspace genutzt, welcher die Worldspace Position von der 
		// Lichtquele wegbewegt. Dadurch bleibt der Peter-Paning Fehler gleich, unabhngig davon wo sich das Objekt befindet und wie 
		// gro der Bereich im Viewfrustum der Lichtquele ist.
		// Abhngig von dem Winkel zwischen Normalen und Lichtrichtung wird ein grerer Offset fr die Schattenberechnung genommen. 
		// Bei steileren Winkeln gibt es mehr Shadow-Acne, also greren Offset nehmen.
		float3 normal = normalize(p_psIn.NormalAndWaterFog.xyz);
		float d = dot(normal, -LightDir);		

		// Falls der Winkel grer als ca 96 Grad ist, numm einen Offset von 40, falls der Winkel kleiner als ca 84 Grad ist,
		// nimm einen Offset von 25. Sonst interpoliere linear.
		d += 0.1f;
		d *= 5.0f;
		d = saturate(d);
		float offset = lerp(25.0f, 40.0f, d);

		// Position um den Offset von der Lichtquele wegbewegen
		float4 worldPos = p_psIn.LightVPS;
		worldPos.xyz -= normalize(LightDir) * offset;

		// Shadowmapping durchfhren.
		float4 v = PerformTransformInLightSpace(worldPos);

		float fCurrentDepth = GetDepthInLightSpace(v);
		fShadow = GetObjectInShadowFactor(fCurrentDepth, v, ShadowMap);
	}
	else
	{
		// by mad: um schatten erweitert
		float fCurrentDepth = GetDepthInLightSpace(p_psIn.LightVPS);
		fShadow = GetObjectInShadowFactor(fCurrentDepth, p_psIn.LightVPS, ShadowMap);
	}
    
    col.rgb *= saturate(p_psIn.Amb + fShadow*p_psIn.Diff) + fShadow*colSpec;
#else
    col.rgb *= saturate(p_psIn.Amb + p_psIn.Diff) + colSpec;
    //return col * p_psIn.Col + p_psIn.Spec;
#endif
	
	if (bUnderWaterFx)
    {
        p_psIn.NormalAndWaterFog.w = saturate(p_psIn.NormalAndWaterFog.w);
        col.rgb += GetCausticColor(p_psIn.NormalAndWaterFog.w, p_psIn.Tex01.xy, p_psIn.Spec.w);
        col = float4LerpXYZ(col, WATERGROUND_COLOR, p_psIn.NormalAndWaterFog.w);
    }

	output.Color = col;

#if defined(USE_DEPTH_MAP_MRT)
	output.Depth.x = 1.0f - p_psIn.PosVS.w;	// Projected depth
	output.Depth.y = p_psIn.PosVS.z;		// Linear view space depth
	output.Depth.a = 1.0f;
#endif

	return output;
}

// ================================================================================================
PS_Out PS_ColorAndTextureAlpha(VS_OutColTex p_psIn, uniform bool bSecondColorBlend = false)
{
	PS_Out output = (PS_Out)0;

#if defined(ENABLE_SHADOWMAPPING)
    // by mad: um schatten erweitert
    float fCurrentDepth = GetDepthInLightSpace(p_psIn.LightVPS);
    float fShadow = GetObjectInShadowFactor(fCurrentDepth, p_psIn.LightVPS, ShadowMap);
#endif
    
    float4 colTex = tex2D(SamplerBase, p_psIn.Tex);
    if (bSecondColorBlend)
    {
        float4 col2 =  tex2D(Sampler4, p_psIn.Tex.xy);
        colTex.a   = ((col2.a + 0.001) >= g_fSecondColorDecay) ? colTex.a : 0;
    }
    
#if defined(ENABLE_SHADOWMAPPING)
    float4 col = clamp(p_psIn.Amb + fShadow * p_psIn.Diff,0,1);
#else
    float4 col = clamp(p_psIn.Amb + p_psIn.Diff,0,1);
#endif
    
    col.a = colTex.a;

	output.Color = col;

    return output;
}

#endif
